@@ -6,7 +6,6 @@ from warnings import warn
 
 import numpy as np
 from scipy.optimize import least_squares
-from damask_parse import read_table as read_damask_table
 
 from formable.maths_utils import get_principal_values
 from formable.utils import requires, at_most_one_of
@@ -108,11 +107,12 @@ class LoadResponse(object):
 
         Returns
         -------
-        yield_stress : ndarray of shape (M, 3, 3)
-            Yield stress tensors, one for each of M yield point criteria values.
-        good_idx : list of int
-            Indices of yield point criteria values for which the yield stress was
-            successfully calculated.
+        Two-tuple of:
+            yield_stress : ndarray of shape (M, 3, 3)
+                Yield stress tensors, one for each of M yield point criteria values.
+            good_idx : list of int
+                Indices of yield point criteria values for which the yield stress was
+                successfully calculated.
 
         """
 
@@ -256,7 +256,7 @@ class LoadResponseSet(object):
                     if not isinstance(uniaxial_response, LoadResponse):
                         raise TypeError('`uniaxial_response` must be a LoadResponse')
 
-                    if not uniaxial_response.is_uniaxial(tol=1e-3):
+                    if not uniaxial_response.is_uniaxial(tol=5e-3):
                         msg = ('Specified `uniaxial_response` does not appear to be '
                                'uniaxial.')
                         raise ValueError(msg)
@@ -325,6 +325,11 @@ class LoadResponseSet(object):
         elif not issubclass(yield_function, YieldFunction):
             raise TypeError(msg)
 
+        bad_kwargs = list(set(kwargs.keys()) - set(yield_function.PARAMETERS))
+        if bad_kwargs:
+            bad_kwargs_fmt = ', '.join([f'"{i}"' for i in bad_kwargs])
+            raise ValueError(f'Unknown yield function parameters: {bad_kwargs_fmt}')
+
         if not yield_point_criteria_idx:
             yield_point_criteria_idx = [
                 [ypc_idx, ypc_val_idx]
@@ -381,7 +386,8 @@ class LoadResponseSet(object):
 
     def show_yield_functions_3D(self, normalise=True, resolution=DEF_3D_RES,
                                 equivalent_stress=None, min_stress=None, max_stress=None,
-                                show_axes=True, planes=None, backend='plotly'):
+                                show_axes=True, planes=None, backend='plotly',
+                                join_stress_states=False, show_contour_grid=False):
         'Visualise all fitted yield functions and data in 3D.'
 
         if not self.yield_functions:
@@ -389,11 +395,15 @@ class LoadResponseSet(object):
 
         yld_funcs = []
         yld_stresses = []
+        stress_indices = []
         for yld_func_dict in self.yield_functions:
             yld_funcs.append(yld_func_dict['yield_function'])
-            yld_stress = self.yield_stresses[yld_func_dict['yield_stress_idx']]['values']
-            yld_stress_principal = get_principal_values(yld_stress)
+            yld_stress = self.yield_stresses[yld_func_dict['yield_stress_idx']]
+            yld_stress_vals = yld_stress['values']
+            resp_idx = yld_stress['response_idx']
+            yld_stress_principal = get_principal_values(yld_stress_vals)
             yld_stresses.append(yld_stress_principal)
+            stress_indices.append(resp_idx)
 
         return YieldFunction.compare_3D(
             yld_funcs,
@@ -405,12 +415,16 @@ class LoadResponseSet(object):
             show_axes=show_axes,
             planes=planes,
             stress_states=yld_stresses,
+            stress_indices=stress_indices,
             backend=backend,
+            join_stress_states=join_stress_states,
+            show_contour_grid=show_contour_grid,
         )
 
     def show_yield_functions_2D(self, plane, normalise=True, resolution=DEF_2D_RES,
                                 equivalent_stress=None, min_stress=None, max_stress=None,
-                                show_axes=True, up=None, show_contour_grid=False):
+                                show_axes=True, up=None, show_contour_grid=False,
+                                join_stress_states=False):
         'Visualise all fitted yield functions and data in 2D.'
 
         if not self.yield_functions:
@@ -418,11 +432,15 @@ class LoadResponseSet(object):
 
         yld_funcs = []
         yld_stresses = []
+        stress_indices = []
         for yld_func_dict in self.yield_functions:
             yld_funcs.append(yld_func_dict['yield_function'])
-            yld_stress = self.yield_stresses[yld_func_dict['yield_stress_idx']]['values']
-            yld_stress_principal = get_principal_values(yld_stress)
+            yld_stress = self.yield_stresses[yld_func_dict['yield_stress_idx']]
+            yld_stress_vals = yld_stress['values']
+            resp_idx = yld_stress['response_idx']
+            yld_stress_principal = get_principal_values(yld_stress_vals)
             yld_stresses.append(yld_stress_principal)
+            stress_indices.append(resp_idx)
 
         return YieldFunction.compare_2D(
             yld_funcs,
@@ -433,6 +451,8 @@ class LoadResponseSet(object):
             min_stress=min_stress,
             max_stress=max_stress,
             stress_states=yld_stresses,
+            stress_indices=stress_indices,
             up=up,
             show_contour_grid=show_contour_grid,
+            join_stress_states=join_stress_states,
         )
