@@ -9,10 +9,9 @@ import abc
 import functools
 
 import numpy as np
-import pyvista as pv
 from plotly import graph_objects
 from plotly.colors import DEFAULT_PLOTLY_COLORS
-from scipy.optimize import least_squares
+from scipy.optimize import least_squares, curve_fit
 
 from formable import utils, maths_utils
 
@@ -89,15 +88,17 @@ class YieldFunction(metaclass=abc.ABCMeta):
         return value
 
     @classmethod
-    def from_fit(cls, stress_states, initial_params=None, **kwargs):
+    def from_fit(cls, stress_states, initial_params=None, force_fit=False, **kwargs):
         """Fit the yield function to yield stress states.
 
         Parameters
         ----------
         stress_states : ndarray of shape (N, 3, 3)
-        initial_params : dict 
+        initial_params : dict
             Any initial guesses for the fitting parameters. Mutually exclusive with
             additional keyword arguments passed, which are considered to be fixed.
+        force_fit : bool, optional
+            If False, do not fit if there is insufficient input data. False by default.
         kwargs : dict
             Additional parameters passed to this method will be fixed during the fit.
 
@@ -113,6 +114,9 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
         initial_params = initial_params or {}
 
+        # print(f'fitting_param_names: {fitting_param_names}')
+        # print(f'initial_params: {initial_params}')
+
         if initial_params:
             for k, initial_guess in initial_params.items():
                 if k not in fitting_param_names:
@@ -131,7 +135,7 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
         else:
 
-            if len(fitting_param_names) > len(stress_states):
+            if len(fitting_param_names) > len(stress_states) and not force_fit:
                 msg = (f'Insufficient number of stress states ({len(stress_states)}) to '
                        f'fit yield function "{cls.__name__}" with '
                        f'{len(fitting_param_names)} fitting parameters.')
@@ -144,9 +148,25 @@ class YieldFunction(metaclass=abc.ABCMeta):
                 idx = fitting_param_names.index('equivalent_stress')
                 initial_params_all[idx] = 50e6
 
+            if 'exponent' in fitting_param_names:
+                bounds = (
+                    [-2] * (len(initial_params_all) - 1) + [1],
+                    [+2] * (len(initial_params_all) - 1) + [8],
+                )
+            else:
+                bounds = (
+                    [-2] * len(initial_params_all),
+                    [+2] * len(initial_params_all),
+                )
+
+            # print(f'initial_params_all: {initial_params_all}')
+            # print(f'fitting_param_names: {fitting_param_names}')
+            # print(f'bounds:\n{bounds}')
+
             fit = least_squares(
                 cls.residual,
                 initial_params_all,
+                bounds=bounds,
                 kwargs=dict(
                     stress_states=stress_states,
                     fitting_param_names=fitting_param_names,
@@ -202,7 +222,9 @@ class YieldFunction(metaclass=abc.ABCMeta):
     @classmethod
     def compare_3D(cls, yield_functions, normalise=True, resolution=DEF_3D_RES,
                    equivalent_stress=None, min_stress=None, max_stress=None,
-                   show_axes=True, planes=None, stress_states=None, backend='plotly'):
+                   show_axes=True, planes=None, stress_states=None, backend='plotly',
+                   stress_indices=None, join_stress_states=False, show_contour_grid=False,
+                   legend_text=None):
         """Visualise one or more yield functions in 3D.
 
         Parameters
@@ -224,6 +246,23 @@ class YieldFunction(metaclass=abc.ABCMeta):
                        'to the `yield_functions` list.')
                 raise ValueError(msg)
 
+            if join_stress_states and not stress_indices:
+                msg = ('If `join_stress_states=True`, `stress_indices` must be specified '
+                       'as a list (of equal length to `stress_states`) of integer '
+                       'indices that determine how the stress states should be '
+                       'connected.')
+                raise ValueError(msg)
+
+            if stress_indices and len(stress_indices) != len(stress_states):
+                msg = '`stress_indices` must have the same length as `stress_states`.'
+                raise ValueError(msg)
+
+        if legend_text:
+            if len(legend_text) != len(yield_functions):
+                msg = ('If specified, `legend_text` must be a list of equal length to '
+                       '`yield_function`.')
+                raise ValueError(msg)
+
         stress_range = cls._get_multi_plot_stress_range(
             yield_functions=yield_functions,
             normalise=normalise,
@@ -248,21 +287,6 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
             raise NotImplementedError('not yet.')
 
-            grid = pv.StructuredGrid(stress_X, stress_Y, stress_Z)
-            for yld_func, i in zip(yield_functions, values_all):
-                grid[yld_func.__class__.__name__] = i
-
-            # TODO: fix this, need a contours object for each yield surface?
-            contours = grid.contour([0])
-
-            pv.set_plot_theme('document')
-            p = pv.PlotterITK()
-            p.add_mesh(contours, scalars=contours.points[:, 2])
-            if stress_states is not None:
-                p.add_points(stress_states)
-
-            return p
-
         elif backend == 'plotly':
 
             stress_X = stress_X.flatten()
@@ -271,10 +295,64 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
             fig_data = []
 
+            if show_contour_grid:
+
+                min_x, max_x = min(stress_X), max(stress_X)
+                min_y, max_y = min(stress_Y), max(stress_Y)
+                min_z, max_z = min(stress_Z), max(stress_Z)
+
+                fig_data.append({
+                    'type': 'scatter3d',
+                    'mode': 'markers',
+                    'marker': {
+                        'symbol': 'x',
+                        'color': 'black',
+                        'size': 4,
+                    },
+                    'x': [min_x, min_x, min_x, min_x, max_x, max_x, max_x, max_x],
+                    'y': [min_y, min_y, max_y, max_y, min_y, min_y, max_y, max_y],
+                    'z': [min_z, max_z, min_z, max_z, min_z, max_z, min_z, max_z],
+                })
+
             for idx, (yld_func, i) in enumerate(zip(yield_functions, values_all)):
 
+                # print(f'i.shape: {i.shape}')
+                # print(f'i: {i}')
+
+                # valid_idx = np.isfinite(i)
+
+                # print(f'valid_idx len: {valid_idx.sum()}')
+
+                # i_valid = i[valid_idx]
+                # stress_X_valid = stress_X[valid_idx]
+                # stress_Y_valid = stress_Y[valid_idx]
+                # stress_Z_valid = stress_Z[valid_idx]
+
+                # print(f'i.shape: {i_valid.shape}')
+                # print(f'i: {i}')
+
+                # print(f'max(i_valid): {max(i_valid)}, min(i_valid): {min(i_valid)}')
+
+                # print(f'i_valid[:100]:\n{i_valid[:100]!r}')
+                # print(f'stress_X_valid[:100]:\n{stress_X_valid[:100]!r}')
+                # print(f'stress_Y_valid[:100]:\n{stress_Y_valid[:100]!r}')
+                # print(f'stress_Z_valid[:100]:\n{stress_Z_valid[:100]!r}')
+
+                near_zero_idx = np.where(np.abs(i) < 1e-3)
+                #print(f'near_zero_idx: {near_zero_idx}')
+
+                near_zero_stress = np.array([
+                    stress_X[near_zero_idx],
+                    stress_Y[near_zero_idx],
+                    stress_Z[near_zero_idx],
+                ]).T
+
+                #print(f'near_zero_stress: {near_zero_stress.shape}\n{near_zero_stress!r}')
+
                 name = f'{idx + 1}. {yld_func.__class__.__name__}'
                 yield_point = getattr(yld_func, 'yield_point', None)
+                if legend_text:
+                    name += f' [{legend_text[idx]}] '
                 if yield_point:
                     name += f' ({yield_point})'
 
@@ -284,19 +362,37 @@ class YieldFunction(metaclass=abc.ABCMeta):
                     'y': stress_Y,
                     'z': stress_Z,
                     'value': i,
+                    'name': name,
+                    'opacity': 0.5,
+                    'showlegend': True,
+                    # ----
                     'isomin': 0,
                     'isomax': 0,
                     'surface_count': 1,
-                    'opacity': 0.5,
                     'colorscale': [
                         [0, DEFAULT_PLOTLY_COLORS[idx]],
                         [1, DEFAULT_PLOTLY_COLORS[idx]],
                     ],
                     'showscale': False,
-                    'showlegend': True,
-                    'name': name,
+                    # ----
+                    # 'surface_count': 4,
+                    # 'showscale': True,
                 })
 
+                # fig_data.append({
+                #     'type': 'scatter3d',
+                #     'mode': 'markers',
+                #     'x': stress_X,
+                #     'y': stress_Y,
+                #     'z': stress_Z,
+                #     'marker': {
+                #         'color': i,
+                #         'showscale': True,
+                #     },
+                #     'name': name,
+                #     'opacity': 0.5,
+                # })
+
             fig_annots = []
 
             if planes:
@@ -336,9 +432,10 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
             if stress_states is not None:
 
-                for idx, (yld_func, stresses) in enumerate(zip(yield_functions, stress_states)):
+                for idx, stresses in enumerate(stress_states):
 
                     if stresses is not None:
+                        stresses = stresses.copy()
 
                         if normalise:
                             stresses /= eq_stress
@@ -357,6 +454,44 @@ class YieldFunction(metaclass=abc.ABCMeta):
                             }
                         )
 
+                if join_stress_states:
+                    max_idx = np.max(np.concatenate(stress_indices))
+                    all_idx = [utils.expand_idx_array(i, max_idx) for i in stress_indices]
+
+                    for i_idx, i in enumerate(all_idx[1:], 1):
+
+                        if (stress_states[i_idx] is None or
+                                stress_states[i_idx - 1] is None):
+                            continue
+
+                        A_exp, A_idx = i
+                        B_exp, B_idx = all_idx[i_idx - 1]
+
+                        matching_idx = np.ma.filled(A_exp == B_exp, fill_value=False)
+                        stress_A_idx = A_idx.data[matching_idx]
+                        stress_B_idx = B_idx.data[matching_idx]
+
+                        A = stress_states[i_idx][stress_A_idx]
+                        B = stress_states[i_idx - 1][stress_B_idx]
+
+                        # Lines should go from A to B
+                        xyz = np.zeros((3, A.shape[0] * 3))
+                        xyz[:, 2::3] = np.nan
+                        xyz[:, 0::3] = A.T
+                        xyz[:, 1::3] = B.T
+
+                        fig_data.append({
+                            'type': 'scatter3d',
+                            'x': xyz[0],
+                            'y': xyz[1],
+                            'z': xyz[2],
+                            'mode': 'lines',
+                            'name': f'Stress path ({i_idx})',
+                            'line': {
+                                'color': 'black',
+                            },
+                        })
+
             fig = graph_objects.FigureWidget(
                 data=fig_data,
                 layout={
@@ -391,8 +526,17 @@ class YieldFunction(metaclass=abc.ABCMeta):
     @classmethod
     def compare_2D(cls, yield_functions, plane, normalise=True, resolution=DEF_2D_RES,
                    equivalent_stress=None, min_stress=None, max_stress=None,
-                   stress_states=None, up=None, show_contour_grid=False):
-        'Visualise multiple yield functions in 2D.'
+                   stress_states=None, up=None, show_contour_grid=False,
+                   stress_indices=None, join_stress_states=False, legend_text=None):
+        """Visualise multiple yield functions in 2D.
+
+        Parameters
+        ----------
+        join_stress_states : bool, optional
+            If True, stress states corresponding to the same `stress_indices` values will
+            be joined by straight lines. False by default.
+
+        """
 
         if stress_states is not None:
             if len(stress_states) != len(yield_functions):
@@ -400,6 +544,23 @@ class YieldFunction(metaclass=abc.ABCMeta):
                        'to the `yield_functions` list.')
                 raise ValueError(msg)
 
+            if join_stress_states and not stress_indices:
+                msg = ('If `join_stress_states=True`, `stress_indices` must be specified '
+                       'as a list (of equal length to `stress_states`) of integer '
+                       'indices that determine how the stress states should be '
+                       'connected.')
+                raise ValueError(msg)
+
+            if stress_indices and len(stress_indices) != len(stress_states):
+                msg = '`stress_indices` must have the same length as `stress_states`.'
+                raise ValueError(msg)
+
+        if legend_text:
+            if len(legend_text) != len(yield_functions):
+                msg = ('If specified, `legend_text` must be a list of equal length to '
+                       '`yield_function`.')
+                raise ValueError(msg)
+
         stress_range = cls._get_multi_plot_stress_range(
             yield_functions=yield_functions,
             normalise=normalise,
@@ -432,6 +593,9 @@ class YieldFunction(metaclass=abc.ABCMeta):
         for idx, (yld_func, i) in enumerate(zip(yield_functions, values_all)):
 
             name = f'{idx + 1}. {yld_func.__class__.__name__}'
+            if legend_text:
+                name += f' [{legend_text[idx]}] '
+
             yield_point = getattr(yld_func, 'yield_point', None)
             if yield_point:
                 name += f' ({yield_point})'
@@ -448,7 +612,8 @@ class YieldFunction(metaclass=abc.ABCMeta):
                     'coloring': 'none',
                 },
                 'line': {
-                    # Resolution must be high enough to accurately portray curvature.
+                    # Resolution must be high enough to accurately portray curvature;
+                    # so don't rely on smoothing!
                     'smoothing': 0,
                     'color': DEFAULT_PLOTLY_COLORS[idx],
                 },
@@ -460,10 +625,10 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
         if stress_states is not None:
 
-            for idx, (yld_func, stresses) in enumerate(zip(yield_functions, stress_states)):
-
+            proj_stresses = []  # 1 element for each yield_function/stress_state
+            for stresses in stress_states:
                 if stresses is not None:
-
+                    stresses = stresses.copy()
                     if normalise:
                         stresses /= eq_stress
 
@@ -484,15 +649,24 @@ class YieldFunction(metaclass=abc.ABCMeta):
                     if not normalise:
                         proj_rot[0:2] /= 1e6
 
-                    color_dim = np.abs(dist) / np.max(np.abs(dist))
+                    proj_stresses.append({
+                        'x': proj_rot[0],
+                        'y': proj_rot[1],
+                        'dist': dist
+                    })
+                else:
+                    proj_stresses.append(None)
 
+            for idx, proj in enumerate(proj_stresses):
+                if proj is not None:
+                    color_dim = np.abs(proj['dist']) / np.max(np.abs(proj['dist']))
                     base_col = DEFAULT_PLOTLY_COLORS[idx]
                     base_col_rgb = utils.parse_rgb_str(base_col)
 
                     fig_data.append({
                         'type': 'scatter',
-                        'x': proj_rot[0],
-                        'y': proj_rot[1],
+                        'x': proj['x'],
+                        'y': proj['y'],
                         'mode': 'markers',
                         'name': f'{idx+ 1}. Stress data',
                         'hovertext': dist,
@@ -507,13 +681,77 @@ class YieldFunction(metaclass=abc.ABCMeta):
                         },
                     })
 
+            if join_stress_states:
+                max_idx = np.max(np.concatenate(stress_indices))
+                all_idx = [utils.expand_idx_array(i, max_idx) for i in stress_indices]
+
+                for i_idx, i in enumerate(all_idx[1:], 1):
+
+                    if proj_stresses[i_idx] is None or proj_stresses[i_idx - 1] is None:
+                        continue
+
+                    A_exp, A_idx = i
+                    B_exp, B_idx = all_idx[i_idx - 1]
+
+                    matching_idx = np.ma.filled(A_exp == B_exp, fill_value=False)
+                    stress_A_idx = A_idx.data[matching_idx]
+                    stress_B_idx = B_idx.data[matching_idx]
+
+                    A_x = proj_stresses[i_idx]['x'][stress_A_idx]
+                    A_y = proj_stresses[i_idx]['y'][stress_A_idx]
+                    B_x = proj_stresses[i_idx - 1]['x'][stress_B_idx]
+                    B_y = proj_stresses[i_idx - 1]['y'][stress_B_idx]
+
+                    # Lines should go from A to B
+                    xy = np.zeros((2, A_x.size * 3))
+                    xy[:, 2::3] = np.nan
+                    xy[0, 0::3] = A_x
+                    xy[0, 1::3] = B_x
+                    xy[1, 0::3] = A_y
+                    xy[1, 1::3] = B_y
+
+                    # Plotly doesn't support varying line color, but if it did, we would
+                    # do this:
+                    avg_dist = 0.5 * (
+                        proj_stresses[i_idx]['dist'][stress_A_idx] +
+                        proj_stresses[i_idx - 1]['dist'][stress_B_idx]
+                    )
+                    color_dim = np.abs(avg_dist) / np.max(np.abs(avg_dist))
+
+                    fig_data.append({
+                        'type': 'scatter',
+                        'x': xy[0],
+                        'y': xy[1],
+                        'mode': 'lines',
+                        'name': f'Stress path ({i_idx})',
+                        'line': {
+                            'width': 0.4,
+                            'color': 'rgba(120, 120, 120, 0.5)',
+                        },
+                    })
+
         if show_contour_grid:
+
+            cg_text_strs = []
+            for idx, (yld_func, i) in enumerate(zip(yield_functions, values_all)):
+                cg_text_i = f'{idx + 1}. {yld_func.__class__.__name__}'
+                if legend_text:
+                    cg_text_i += f' [{legend_text[idx]}] '
+                cg_text_strs.append(cg_text_i)
+
+            cg_text_all = []
+            for cg_text_i, i in zip(cg_text_strs, values_all):
+                cg_text_all.append([f'{cg_text_i}: {j:.3f}' for j in i])
+
+            cg_text_final = ['<br>'.join(list(i)) for i in zip(*cg_text_all)]
+
             fig_data.append(
                 {
                     'type': 'scatter',
                     'x': grid_coords_2D[0],
                     'y': grid_coords_2D[1],
                     'mode': 'markers',
+                    'text': cg_text_final,
                     'marker': {
                         'size': 3,
                         'color': 'gray',
@@ -796,16 +1034,13 @@ class YieldFunction(metaclass=abc.ABCMeta):
             resolution=resolution,
         )
 
-        # TODO: sort this out:
-        # Remove any NaN values:
-        # valid_idx = np.logical_not(np.isnan(grid_values))
-        # grid_values = grid_values[valid_idx]
-        # stress_grid = stress_grid[valid_idx]
-
         grid_values_all = []
         for yld_func in yield_functions:
             values_i = yld_func.get_value(stress_grid * equivalent_stress)
-            values_i /= equivalent_stress
+            # Normalise values solely for the purpose of the visualisation (so contour
+            # values are of the order 1):
+            abs_max = np.nanmax(np.abs(values_i))
+            values_i /= abs_max
             grid_values_all.append(values_i)
 
         return (stress_X, stress_Y, stress_Z), grid_values_all
@@ -848,7 +1083,10 @@ class YieldFunction(metaclass=abc.ABCMeta):
         grid_values_all = []
         for yld_func in yield_functions:
             values_i = yld_func.get_value(stress_grid * equivalent_stress)
-            values_i /= equivalent_stress
+            # Normalise values solely for the purpose of the visualisation (so contour
+            # values are of the order 1):
+            abs_max = np.nanmax(np.abs(values_i))
+            values_i /= abs_max
             grid_values_all.append(values_i)
 
         return grid_coords_2D, grid_values_all, basis_unit
@@ -908,7 +1146,8 @@ class YieldFunction(metaclass=abc.ABCMeta):
 
     def show_3D(self, normalise=True, resolution=DEF_3D_RES, equivalent_stress=None,
                 min_stress=None, max_stress=None, show_axes=True, planes=None,
-                stress_states=None, backend='plotly'):
+                stress_states=None, backend='plotly', stress_indices=None,
+                join_stress_states=False, show_contour_grid=False, legend_text=None):
         """
         Parameters
         ----------
@@ -927,11 +1166,16 @@ class YieldFunction(metaclass=abc.ABCMeta):
             planes=planes,
             stress_states=stress_states,
             backend=backend,
+            stress_indices=stress_indices,
+            join_stress_states=join_stress_states,
+            show_contour_grid=show_contour_grid,
+            legend_text=legend_text,
         )
 
     def show_2D(self, plane, normalise=True, resolution=DEF_2D_RES,
                 equivalent_stress=None, min_stress=None, max_stress=None,
-                stress_states=None, up=None, show_contour_grid=False):
+                stress_states=None, up=None, show_contour_grid=False, stress_indices=None,
+                join_stress_states=False, legend_text=None):
         """
         Parameters
         ----------
@@ -950,4 +1194,83 @@ class YieldFunction(metaclass=abc.ABCMeta):
             stress_states=stress_states,
             up=up,
             show_contour_grid=show_contour_grid,
+            stress_indices=stress_indices,
+            join_stress_states=join_stress_states,
+            legend_text=legend_text,
         )
+
+    def get_numerical_lankford(self, fit_domain=0.05, resolution=100, tol=2e-2):
+        """Get the Lankford coefficient by numerically fitting the tangent to the yield
+        function.
+
+        Parameters
+        ----------
+        fit_domain : float, optional
+            The side lengths of the box within which to fit a line (i.e. the yield
+            function tangent), expressed as a fraction of the equivalent stress.
+        resolution : int, optional
+            Size of the grid within the fitting domain, on which values of the yield
+            function will be evaluated.
+        tol : float, optional
+            Yield function values that are within this value from zero will be included
+            in the tangent fit.
+
+        Returns
+        -------
+        lankford : float
+
+        """
+
+        if not hasattr(self, 'equivalent_stress'):
+            msg = 'Yield function must have an `equivalent_stress` parameter.'
+            raise NotImplementedError(msg)
+        else:
+            equiv_stress = getattr(self, 'equivalent_stress')
+
+        X, Y = np.meshgrid(*[np.linspace(-0.5, 0.5, num=resolution)] * 2)
+        xy = np.vstack([X.flatten(), Y.flatten()])
+        xy *= equiv_stress * fit_domain
+        xy += np.array([[equiv_stress, 0]]).T
+
+        num_ss = xy.shape[1]
+        stress_states = np.tile(np.eye(3), (num_ss, 1)).reshape(num_ss, 3, 3)
+        stress_states[:, 0, 0] = xy[0]
+        stress_states[:, 1, 1] = xy[1]
+
+        vals = self.get_value(stress_states)
+
+        # print(f'vals:\n{vals}\n')
+        good_idx = np.where((np.abs(vals) - tol) < 0)[0]
+        if not good_idx.size:
+            raise ValueError(f'No yield surface solutions within `tol={tol}`.')
+
+        # print(f'good_idx:\n{good_idx}\n')
+
+        # [# For fitting, normalise x, y values by equivalent stress (we just want the
+        # # direction of the tangent vector in the σ_3 = 0 plane)] ?
+        # xy_norm = xy / equiv_stress
+
+        fit_coords = xy[:, good_idx]
+        # Get x, y values to fit:
+
+        # print(f'fit_coords:\n{fit_coords}\n')
+
+        popt, pcov = curve_fit(maths_utils.line, fit_coords[0], fit_coords[1])
+        # print(f'popt: {popt}\n; pcov:\n{pcov}\n')
+
+        m, c = popt
+        if c < 0:
+            m, c = -m, -c
+
+        tangent_x = [np.min(xy[0]), np.max(xy[0])]
+        tangent_y = maths_utils.line(np.array(tangent_x), m, c)
+
+        normal = np.array([-m, 1])
+        # print(f'normal: {normal}')
+
+        normal_n = normal / np.linalg.norm(normal)
+        # print(f'normal_n: {normal_n}')
+
+        lankford = normal_n[1] / (-normal[0] - normal[1])
+
+        return (lankford, (tangent_x, tangent_y))
